/*

kHTTPd 0.1.0pre1

(C) 1999 by Arjan van de Ven

*/

/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/


#include <linux/module.h>
#include <linux/kernel.h>

#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/net.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/smp_lock.h>
#include <linux/un.h>
#include <linux/unistd.h>

#include <net/ip.h>
#include <net/sock.h>
#include <net/tcp.h>

#include <asm/atomic.h>
#include <asm/semaphore.h>
#include <asm/processor.h>
#include <asm/uaccess.h>

#include <linux/file.h>


#include "khttpd.h"
#include "khttpd_mime.h"

/* Two atomic counters for child-management */

atomic_t 	AcceptCount; 	/* The number of children in "accept()" */
atomic_t	InitCount;    	/* The number of children that is newly
				   created but isn't yet in accept */ 

int errno; /* FIXME: This seems to be required for waitpid, but I don't like it! */


int khttpd_pid;

/*

Sendbuffer sends "Length" bytes from Buffer to the socket "sock".

*/

size_t SendBuffer(struct socket *sock, const char *Buffer, size_t Length)
{
	struct msghdr	msg;
	mm_segment_t	oldfs;
	struct iovec	iov;
	int 		len,len2;
	
	DebugStr("Entering SendBuffer\n");
	
	
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov;
	msg.msg_iovlen   = 1;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = 0/*MSG_DONTWAIT*/;    
	iov.iov_base = (char*) Buffer;
	iov.iov_len  = (size_t)Length;
	
	
	len = 0;
/*	while (len<Length)*/
	{
		oldfs = get_fs(); set_fs(KERNEL_DS);
		len2 = sock_sendmsg(sock,&msg,(size_t)(Length-len));
		set_fs(oldfs);
		if ((len+len2)<Length) 
		{
			schedule();
		}
		if (len2>0) len+=len2;		
		
/*		if ((len2<0)&&(len2!=-EAGAIN)) break;*/
	}

	return len;	
}

/* 
   SendBuffer2 send TWO buffers. This is used when sending the header along
   with the first data-block. The later is imperative in order to achieve
   an acceptable performance. See the apache-performance document.
*/
size_t SendBuffer2(struct socket *sock, const char *Buffer1, size_t Length1,
	 	   const char *Buffer2, size_t Length2)
{
	struct msghdr	msg;
	mm_segment_t	oldfs;
	struct iovec	iov[2];
	int 		len,len2,Length;
	
	DebugStr("Entering SendBuffer2\n"); 
	
	
	Length 		 = Length1 + Length2;
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov[0];
	msg.msg_iovlen   = 2;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = 0 /*MSG_DONTWAIT*/; 
	
	iov[0].iov_base = (char *)Buffer1;
	iov[0].iov_len  = (size_t)Length1;
	iov[1].iov_base = (char *)Buffer2;
	iov[1].iov_len  = (size_t)Length2;
	
	len = 0;
/*	while (len<Length)*/
	{
		oldfs = get_fs(); set_fs(KERNEL_DS);
		len2 = sock_sendmsg(sock,&msg,(size_t)(Length1+Length2-len));
		set_fs(oldfs);
		if ((len+len2)<Length) 
		{
			schedule();
		}
		if (len2>0) len+=len2;		
		
/*		if ((len2<0)&&(len2!=-EAGAIN)) break;*/
	}

	return len;	
}


/*

khttp_copy_filp_to_sock copies "FileLength" bytes from the file "*filp" to
the socket "sock". This should be optimized to resemble the "sendfile" syscall

return value 0 = success,  <0 = failure
*/
#ifndef HAVE_DO_GENERIC_FILE_READ

int khttp_copy_filp_to_sock(struct file *filp,struct socket *sock,
		size_t FileLength, const char *Header)
{
	int BytesSend=0;
	char *DataBuffer;
#ifdef HAVE_TCP_CORK	
	struct sock *sk=sock->sk;
	struct tcp_opt *tp = &(sk->tp_pinfo.af_tcp);
#endif

	DebugStr("Entering filp2sock\n");
	
	if (!filp->f_op->read)
	 	return -1;
	 	

#ifdef HAVE_TCP_CORK
/*	Set TCP_CORK */
	sk = sock->sk;
	sk->nonagle = 2;	
#endif			
	DataBuffer=(char *)get_free_page(GFP_KERNEL);
	if (DataBuffer==NULL)
	{
		return -1;
	}

	
	/* Start sending the file */
		
	while (BytesSend<FileLength)
	{
		int 		BytesDone;
		int 		ThisPass;
		mm_segment_t	oldfs;
		
		/* Determine the bytes for this pass */
		ThisPass = FileLength-BytesSend;
		if (ThisPass>4096)
			ThisPass = 4096;
			  	
			  
		/* Read a block of "ThisPass" bytes from the file */
		
		filp->f_pos = BytesSend;
		oldfs = get_fs(); set_fs(KERNEL_DS);
		BytesDone = filp->f_op->read(filp, DataBuffer, ThisPass, &filp->f_pos);
		set_fs(oldfs);

	
		if (BytesDone<=0)
		{
			free_page((unsigned long)DataBuffer);
		  	return -1;
	 	}
	 	
	 	/* Send the buffer to the client */
		
		if (!BytesSend)
		SendBuffer2(sock,Header,strlen(Header),DataBuffer,BytesDone);
		else
		SendBuffer(sock,DataBuffer,BytesDone);
			  
		BytesSend+=BytesDone;			 			  
	}

	free_page((unsigned long) DataBuffer);

#ifdef HAVE_TCP_CORK
/* Remove TCP_CORK  */	
	sk->nonagle = 0;
	lock_sock(sk);
	tcp_push_pending_frames(sk,tp);
	release_sock(sk);
#endif	 
	return 0;
}

#else
static int sock_send_actor(read_descriptor_t * desc, const char *area, unsigned long size)
{
	int written;
	unsigned long count = desc->count;
	struct socket *sock = (struct socket *) desc->buf;

	if (size > count)
		size = count;
	written = SendBuffer(sock,(char *)area,size);

	if (written < 0) {
		desc->error = written;
		written = 0;
	}
	desc->count = count - written;
	desc->written += written;
	return written;
}
int khttp_copy_filp_to_sock(struct file * in_file,struct socket *sock, size_t count, const char *Header)
{
	ssize_t retval;
	struct inode * in_inode;
	struct sock *sk=sock->sk;
	struct tcp_opt *tp = &(sk->tp_pinfo.af_tcp);

#ifdef HAVE_TCP_CORK
/*	Set TCP_CORK */
	sk = sock->sk;
	sk->nonagle = 2;	
#endif

	SendBuffer(sock,Header,strlen(Header));
	lock_kernel();

	/*
	 * Get input file, and verify that it is ok..
	 */
	in_inode = in_file->f_dentry->d_inode;

	retval = 0;
	if (count) {
		read_descriptor_t desc;
		loff_t *ppos;

		retval = -EFAULT;
		ppos = &in_file->f_pos;

		desc.written = 0;
		desc.count = count;
		desc.buf = (char *) sock;
		desc.error = 0;
		do_generic_file_read(in_file, ppos, &desc, sock_send_actor);

		retval = desc.written;
		if (!retval)
			retval = desc.error;
	}


	unlock_kernel();
#ifdef HAVE_TCP_CORK	
/* Remove TCP_CORK */	
	sk->nonagle = 0;
	lock_sock(sk);
	tcp_push_pending_frames(sk,tp);
	release_sock(sk);
#endif	
	return retval;
}

#endif
/*

Forwards a connection to userspace.

*/
int HandleUserSpace(struct socket *sock, char *ReceivedSoFar, size_t length)
{
	struct msghdr		msg;
	struct iovec		iov;
	int 			len;
	mm_segment_t		oldfs;
	char			*Buffer;
	
	struct socket		*clientsock;
/*	struct sockaddr_un	name;*/
	struct sockaddr_in 	sin;
	int			error,count=0;
	
	DebugStr("Entering HandleUserSpace\n");	
	
	
	if (ReceivedSoFar==NULL)
	{
		return -500;
	}

	Buffer = (char*) get_free_page(GFP_KERNEL);
	
	if (Buffer==NULL)
	{
		return -500;
	}

#ifdef USE_UNIX	
	error = sock_create(PF_UNIX,SOCK_STREAM,0,&clientsock);
	if (error<0) 
	{
	     printk("Error during creation of client-socket; terminating\n");
	     free_page((unsigned long)Buffer);
	     return -500;
	}



	/* Now bind the socket */
	
	name.sun_family	= AF_UNIX;
	strcpy(name.sun_path,"/tmp/httpd");
	
	size = sizeof(name);

	error = clientsock->ops->connect(clientsock,(struct sockaddr*)&name,size,0);
	if (error<0)
	{
		printk("Error connecting client socket %i \n",error);
		free_page((unsigned long)Buffer);
		return 0;	
	}
#else
	

	/* First create a socket */
	
	error = sock_create(PF_INET,SOCK_STREAM,IPPROTO_TCP,&clientsock);
	if (error<0) 
	     printk("Error during creation of socket; terminating\n");



	/* Now bind the socket */
	
	sin.sin_family	     = AF_INET;
	sin.sin_addr.s_addr  = htonl(INADDR_LOOPBACK);
	sin.sin_port         = htons(KHTTPD_CLIENTPORT);
	

	error = clientsock->ops->connect(clientsock,(struct sockaddr*)&sin,sizeof(sin),0);
	if (error<0)
	{
		printk("Error connecting client socket %i \n",error);
		free_page((unsigned long)Buffer);
		return 0;	
	}
#endif	

	SendBuffer(clientsock,ReceivedSoFar,length);
	len=1;
	
	while ((len>0)&&(count<10000))
	{
		/* First, peek if the remote user has sent some more data */
		msg.msg_name     = 0;
		msg.msg_namelen  = 0;
		msg.msg_iov	 = &iov;
		msg.msg_iovlen   = 1;
		msg.msg_control  = NULL;
		msg.msg_controllen = 0;	
		msg.msg_flags    = 0;
	
		iov.iov_base = &Buffer[0];
		iov.iov_len  = (size_t)4096;
	
		oldfs = get_fs(); set_fs(KERNEL_DS);
		len = sock_recvmsg(sock,&msg,4096,MSG_DONTWAIT);
		set_fs(oldfs);
		
		if ((len!=-EAGAIN)&&(len!=0))
		  printk("From remote : %i bytes \n",len);
	   
		if (len>0)
		  SendBuffer(clientsock,Buffer,len);
		if (len<-100) 
			break;  

		/* Then check if we need to forward something */
		msg.msg_name     = 0;
		msg.msg_namelen  = 0;
		msg.msg_iov	 = &iov;
		msg.msg_iovlen   = 1;
		msg.msg_control  = NULL;
		msg.msg_controllen = 0;	
		msg.msg_flags    = 0;
	
		iov.iov_base = &Buffer[0];
		iov.iov_len  = 4096;
	
		oldfs = get_fs(); set_fs(KERNEL_DS);
		len = sock_recvmsg(clientsock,&msg,4096,MSG_DONTWAIT);
		set_fs(oldfs);
	   
		if (len>0)
		  SendBuffer(sock,Buffer,len);
		 
		if ((len!=-EAGAIN)&&(len<0))
		   printk("From apache: %i \n",len);

		if (sock->sk->state!=TCP_ESTABLISHED)
		   break;
		if (clientsock->sk->state!=TCP_ESTABLISHED)
		   break;

		if ((len<=10)&&(len>-100))
		{
			interruptible_sleep_on_timeout(&clientsock->wait,HZ/20);
		};
		
		if (len==-11) 
		{
			len=1; /* If nothing there, continue the loop */
		}
		count++;
	}
	sock_release(clientsock);

	
	free_page((unsigned long)Buffer);
	return 0;
}


/*
does the actual work for a http-connection, such
as receiving packets and sending the file.

Return value is a http-like value. Positive values are handled by 
this function, negative ones still have to be taken care of. 
*/
int HandleIncommingConnection(struct socket *sock, char *Buffer, size_t *BufLen,
				struct http_time *httptime)
{
	struct msghdr		msg;
	struct iovec		iov;
	int 			len;
	mm_segment_t		oldfs;
	struct http_header 	Head;
	
	DebugStr("Entering HandleIncommingConnection\n");	
	
	
	/* Receive a packet */  /* FIXME: Handle fragemented reads! */
	msg.msg_name     = 0;
	msg.msg_namelen  = 0;
	msg.msg_iov	 = &iov;
	msg.msg_iovlen   = 1;
	msg.msg_control  = NULL;
	msg.msg_controllen = 0;
	msg.msg_flags    = 0;
	
	iov.iov_base = &Buffer[0];
	iov.iov_len  = (size_t)4096;
	
	oldfs = get_fs(); set_fs(KERNEL_DS);
	len = sock_recvmsg(sock,&msg,1024,0);
	set_fs(oldfs);
	
	if (len<0) 
	{
		free_page((unsigned long)Buffer);
		return -500;		
	}
	
	/* Check if the request is finished */
	if ((len<4))
	{
		int len2;
		msg.msg_name     = 0;
		msg.msg_namelen  = 0;
		msg.msg_iov	 = &iov;
		msg.msg_iovlen   = 1;
		msg.msg_control  = NULL;
		msg.msg_controllen = 0;
		msg.msg_flags    = 0;
	
		iov.iov_base = &Buffer[len+1];
		iov.iov_len  = (size_t)4096-len;
	
		oldfs = get_fs(); set_fs(KERNEL_DS);
		len2 = sock_recvmsg(sock,&msg,1024,MSG_DONTWAIT);
		set_fs(oldfs);
		
		interruptible_sleep_on_timeout(&sock->wait,HZ*5);
		
	
		if (len2>0) 
			len+=len2;
	}
	
	Buffer[len+1]=0;

	*BufLen	 = len;
	

	ParseHeader(Buffer,len,&Head);
	
	
	
	if (Head.FileName[0]!=0)
	{
	
		struct file   * filp;
		
		/* Security/Sanity check */
		
	
		if ((strstr(Head.FileName,"..")!=NULL)||
		    (Head.FileName[0]!='/'))
		{
			return -403;
		}
		/*                       */
				
		filp = filp_open(Head.FileName,00,O_RDONLY); /*FIXME: Add permission-check here!*/
		if (IS_ERR(filp))
		{
			return -404;
		}	
		if ((filp!=NULL)&&(filp->f_dentry!=NULL))
		{
			int FileLength,Permission;
			char *Header;
			
			Permission = (int)(filp->f_dentry->d_inode->i_mode)&S_IROTH;			
			if (Permission==0)  /* No permission! */
			{
				fput(filp);
				return -403;
			}
			Permission = (int)(filp->f_dentry->d_inode->i_mode)&S_IFMT;			
			if (Permission==S_IFDIR)  /* Directory */
			{
				fput(filp);
				return -403;
			}
			Permission = (int)(filp->f_dentry->d_inode->i_mode)&(25);			
			if (Permission!=0)  /* Executable */
			{
				fput(filp);
				return -403;
			}
			FileLength = (int)(filp->f_dentry->d_inode->i_size);
			
			if (Head.IMS[0]>0)
			{
				time_t TIME;
				TIME=mimeTime_to_UnixTime(Head.IMS);
				if (TIME>=filp->f_dentry->d_inode->i_mtime)
				{	
					fput(filp);
					Error304(sock);
					return 304;
				}
			}
	

			Header = HTTPHeader(Head.FileName,FileLength,filp->f_dentry->d_inode->i_mtime,httptime);
			
			if (Header==NULL)
			{
				fput(filp);
				return -500;
			}
			khttp_copy_filp_to_sock(filp,sock,FileLength,Header);
			
			if (Header)
			    kfree(Header);
			
			fput(filp);
		}
		

		return 200;		
	}	
	
	return -404;
}


/*

NeedSpawn checks if there are enough children; if not, some more are forked.

*/
 void NeedSpawn(struct socket *sock)
{
	while (atomic_read(&InitCount)+atomic_read(&AcceptCount)<KHTTPD_MINACCEPT)
	{
		kernel_thread(khttpd_child,sock,0);
		atomic_inc(&InitCount);	
	}
}


int khttpd_child(void *TMP)
{
	struct socket *sock,*newsock;
	struct sock *sk;
	int error;
	struct proto_ops *ops;
	char *Buffer;
	size_t BufLen;
	struct http_time  *httptime;
	
	
	DebugStr("Entering khttp_child\n");
	
	current->session = 1;
	current->pgrp    = 1;
	current->state	|= TASK_WAKE_ONE;
	
	sprintf(current->comm,"khttpd - request");
  
	sock = (struct socket*)TMP;

	sk = sock->sk;	
	sk->nonagle = 1;
	sk->linger  = 1;
	
	
	if (sock==NULL) 
	{
		DebugStr("PASSED A NULL POINTER !! \n");
		return 0;
	}

	error=0;
	Buffer = (char*) get_free_page(GFP_KERNEL);
	httptime = kmalloc(sizeof(struct http_time),GFP_KERNEL);
	
	atomic_dec(&InitCount);	
	
	if (Buffer == NULL) return 1;
	if (httptime == NULL)
	{
		free_page((unsigned long)Buffer);
		return 1;
	}
	
	memset(httptime,0,sizeof(struct http_time));
	
	while (1==1)
	{
		/* If there are to many children, break out the loop and terminate */
		if (atomic_read(&AcceptCount)>KHTTPD_MAXACCEPT)
			break;	
	
	
#ifdef DEBUG
		if (signal_pending(current)) 
			break;
#endif		
		/* Before accept: Clone the socket */
		newsock = sock_alloc();
	
		newsock->type = sock->type;
		sock->ops->dup(newsock,sock);
	
		ops = newsock->ops;

		/* Do the actual accept */
		
		atomic_inc(&AcceptCount);

		error = ops->accept(sock,newsock,0);

		atomic_dec(&AcceptCount);
		
		if (error<0)
		{
		 	DebugStr("Error accepting socket\n") ;
		 	break;
		}
		
		/* Do the work */
	
		error=HandleIncommingConnection(newsock,Buffer,&BufLen,httptime);	
		
		if (error<0)
		{
			if (HandleUserSpace(newsock,Buffer,BufLen)<0)
			  ErrorXXX(newsock,-error);
		};
		
	
		/* Remove the cloned socket */
		sock_release(newsock);
		error=1;	
	}




	free_page((unsigned long)Buffer);
	kfree(httptime);
	DebugStr("Leaving khttp_child\n");
	return 0;
}


void KillChildren(pid_t parent_pid)
{
 /*
 
   I don't know how to do this.... yet.
 
 */
}


int khttpdaemon(void *unused)
{

	struct socket *sock;
	struct sockaddr_in sin;
	struct wait_queue *WQ=NULL;
	
	int error; 
	int count,iter;
	
	MOD_INC_USE_COUNT;
	
	sprintf(current->comm,"khttpd - master");
	
	
	current->session = 1;
	current->pgrp    = 1;


	/* First create a socket */
	
	error = sock_create(PF_INET,SOCK_STREAM,IPPROTO_TCP,&sock);
	if (error<0) 
	     printk("Error during creation of socket; terminating\n");



	/* Now bind the socket */
	
	sin.sin_family	     = AF_INET;
	sin.sin_addr.s_addr  = INADDR_ANY;
	sin.sin_port         = htons(KHTTPD_PORT);
	
	error = sock->ops->bind(sock,(struct sockaddr*)&sin,sizeof(sin));
	if (error<0)
	{
		printk("Error binding socket \n");
		MOD_DEC_USE_COUNT;
		return 0;	
	}
	
	sock->sk->reuse   = 1;
	sock->sk->nonagle = 1;
	sock->sk->linger  = 1;

	/* Now, start listening on the socket */
	
	/* I have no idea if "10" is a good backlog-value. */
	
	error=sock->ops->listen(sock,32);	
	if (error!=0)
		printk("Error listening on socket \n");
	sock->flags |= SO_ACCEPTCON;	
	
	/* Now start accepting */
	
	count = 0;  iter=0;

	NeedSpawn(sock);
	UpdateCurrentDate();
	
	/* Wait for any children to die */
	while (1==1)
	{
		int waitpid_result,status;
		sigset_t tmpsig;

		/* Block everything but SIGKILL/SIGSTOP */
		spin_lock_irq(&current->sigmask_lock);
		tmpsig = current->blocked;
		siginitsetinv(&current->blocked, sigmask(SIGKILL) | sigmask(SIGSTOP)| sigmask(SIGHUP));
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);

		waitpid_result = waitpid(-1,NULL, __WCLONE|WNOHANG);

		status=0;

		if (waitpid_result>0) count++;

		/* Allow signals again.. */
		spin_lock_irq(&current->sigmask_lock);
		current->blocked = tmpsig;
		recalc_sigpending(current);
		spin_unlock_irq(&current->sigmask_lock);

		interruptible_sleep_on_timeout(&WQ,KHTTPD_SLEEPVALUE);

		
		if ((iter%KHTTPD_DATEDELAY)==0)		
			UpdateCurrentDate();

#ifdef DEBUG
		if (iter>600)
			atomic_set(&AcceptCount,1000);
		if (iter==601)
			printk("You can kill me soon.\n");
#endif		
		NeedSpawn(sock);
		iter++;
		if (signal_pending(current)) 
			break;
	}
		
	 	



	printk("Ending KHTTP Daemon \n");
	
	atomic_set(&AcceptCount,1000);
	sock->ops->shutdown(sock,2);
	
	KillChildren(khttpd_pid);
	
	interruptible_sleep_on_timeout(&WQ,HZ);


	error = 0;
        while (error>0)
        	error = waitpid(-1,NULL, __WCLONE|WNOHANG);
	
	sock_release(sock);
	
	
	MOD_DEC_USE_COUNT;
	
 	return 0;  
}


int init_module(void)
{
	printk("HTTPD starting up...\n");
	
	atomic_set(&AcceptCount,0);
	atomic_set(&InitCount,0);
 	atomic_set(&MimeCount,0);
 	
 	AddMimeType(".htm","text/html");
 	AddMimeType("html","text/html");
 	AddMimeType(".gif","image/gif");
 	AddMimeType(".jpg","image/gif");
 	AddMimeType(".png","image/gif");
 	AddMimeType(".zip","application/zip");
 	AddMimeType("r.gz","application/x-gtar");
 	AddMimeType(".tgz","application/x-gtar");
	AddMimeType(".tgz","application/x-gtar");
	AddMimeType(".deb","application/x-debian");
	AddMimeType("lass","application/x-java");

  
	khttpd_pid = kernel_thread(khttpdaemon,NULL,0);
  
  
	return 0;
}

void cleanup_module(void)
{
	printk("HTTPD exiting....\n");
}