/*

kHTTPd 0.1.0pre1

(C) 1999 by Arjan van de Ven

*/

/****************************************************************
 *	This program is free software; you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation; either version 2, or (at your option)
 *	any later version.
 *
 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License
 *	along with this program; if not, write to the Free Software
 *	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 ****************************************************************/


#include <linux/kernel.h>

#include <linux/ctype.h>
#include <linux/errno.h>
#include <linux/malloc.h>
#include <linux/net.h>
#include <linux/sched.h>
#include <linux/skbuff.h>
#include <linux/unistd.h>
#include <linux/file.h>

#include <net/ip.h>
#include <net/sock.h>

#include <asm/atomic.h>
#include <asm/semaphore.h>
#include <asm/processor.h>
#include <asm/uaccess.h>



#include "khttpd.h"

atomic_t	MimeCount;

struct MimeType
{
	__u32 	identifier;
	char	type[32-4];  
};

static struct MimeType	MimeTypes[KHTTPD_NUMMIMETYPES];


/*

Send the user-app a "File not found" message

*/
void Error404(struct socket *sock)
{
	 char Message[]=
	"HTTP/1.0 404 File Not Found\r\nServer: KHTTPD/0.0\r\nContent-type: text/html\r\n\r\n<html><body>FILE NOT FOUND!!</body></html>";

	SendBuffer(sock,Message,sizeof(Message));		
	
}

/*
	Access denied 
*/
void Error403(struct socket *sock)
{
	 char Message[]=
	"HTTP/1.0 403 Permission Denied\r\nServer: KHTTPD/0.0\r\nContent-type: text/html\r\n\r\n<html><body>Permission Denied</body></html>";

	SendBuffer(sock,Message,sizeof(Message));		
	
}
/*

Send the user-app a "Not modified"

*/
void Error304(struct socket *sock)
{
	 char Message[]=
	"HTTP/1.0 304 Not modified\r\nServer: KHTTPD/0.0\r\nContent-type: text/html\r\n\r\n";

	SendBuffer(sock,Message,sizeof(Message));		
	
}
void Error500(struct socket *sock)
{
	 char Message[]=
	"HTTP/1.0 304 Service temporarily not available\r\nServer: KHTTPD/0.0\r\nContent-type: text/html\r\n\r\n";

	SendBuffer(sock,Message,sizeof(Message));		
	
}

void ErrorXXX(struct socket *sock, int Error)
{
  if (Error == 404)
  {	
  	Error404(sock);
  	return;
  }
  if (Error == 403)
  {	
  	Error403(sock);
  	return;
  }
  if (Error == 304)
  {	
  	Error304(sock);
  	return;
  }
  if (Error == 500)
  {	
  	Error304(sock);
  	return;
  }
  
  Error404(sock);
}

void AddMimeType(char *Ident,char *Type)
{	
	__u32	*I;
	
	if (strlen(Ident)!=4) 
   	{	
   		printk("httpd: Only 4-byte mime-identifiers are accepted\n");
   		return;
   	}

	if (strlen(Type)>(32-4)) 
   	{	
   		printk("httpd: Mime-string too long.\n");
   		return;
   	}
   	
   	I=(__u32*)Ident;
   	
   	/* FIXME: Need to lock-down all access to the mime-structure here */
   	
   	MimeTypes[atomic_read(&MimeCount)].identifier=*I;
   	strcpy(MimeTypes[atomic_read(&MimeCount)].type,Type);
   	
   	atomic_inc(&MimeCount);
}


char *ResolveMimeType(char *File)
/*

	The returned string is for READ ONLY!

*/
{	
	__u32	*I;
	int pos,lc;
	
	if (strlen(File)<4) 
   	{	
   		return NULL;
   	}

	pos = strlen(File)-4;
   	I=(__u32*)(File+pos);
   	
   	lc=0;
   	
   	while (lc<atomic_read(&MimeCount))
 	{
   		if (MimeTypes[lc].identifier == *I)
   	  		return MimeTypes[lc].type;
   	  	lc++;
   	} 	
   	return MimeTypes[0].type;
}

char *HTTPHeader(char *Filename,int FileSize,time_t MTime, struct http_time *httptime)
{
	char *BUF;
     
	BUF=kmalloc(2048,GFP_KERNEL);
     
        if (MTime!=httptime->UnixTime)
        {
		time_Unix2RFC(MTime,httptime->MimeTime);
        	httptime->UnixTime = MTime;
        };
	sprintf(BUF,
	  "HTTP/1.0 200 OK\r\nServer: KHTTPD/0.0\r\nDate: %s\r\nContent-type: %s\r\nLast-modified: %s\r\nContent-length: %i\r\n\r\n",
	  CurrentTime,ResolveMimeType(Filename),httptime->MimeTime,FileSize);
	

	return BUF;
}



void ParseHeader(char *Buffer,int length, struct http_header *Head)
{
	char *Endval,*EOL,*tmp;
	
	Endval = Buffer + length;
	
	memset(Head,0,sizeof(struct http_header));
	
	while (Buffer<Endval)
	{
		if (isspace(Buffer[0]))
		{
			Buffer++;
			continue;
		};
			
		
		EOL=strchr(Buffer,'\n');
		
		if (EOL==NULL) EOL=Endval;
		
		if (EOL-Buffer<4) 
		{
			Buffer++;
			continue;
		};
		
		if (strncmp("GET ",Buffer,4)==0)
		{
			Buffer+=4;
			
			tmp=strchr(Buffer,' ');
			if (tmp==0) tmp=EOL;
			
			if (tmp>Endval) continue;
			
			strcpy(Head->FileName,KHTTPD_PREFIX);
			strncat(Head->FileName,Buffer,min(255-strlen(KHTTPD_PREFIX),tmp-Buffer));
					
			Buffer=EOL+1;	
			continue;
		}
		if (strncmp("If-Modified-Since: ",Buffer,19)==0)
		{
			Buffer+=19;
			
			strncpy(Head->IMS,Buffer,min(128,EOL-Buffer-1));
					
			Buffer=EOL+1;	
			continue;
		}

		if (strncmp("User-Agent: ",Buffer,12)==0)
		{
			Buffer+=12;
			
			strncpy(Head->Agent,Buffer,min(128,EOL-Buffer-1));
					
			Buffer=EOL+1;	
			continue;
		}
		

		if (strncmp("Host: ",Buffer,6)==0)
		{
			Buffer+=6;
			
			strncpy(Head->Host,Buffer,min(128,EOL-Buffer-1));
					
			Buffer=EOL+1;	
			continue;
		}
		
		Buffer = EOL+1;  /* Skip line */
	}

}